//Added serial TXData
//For PCB V1.0
//This code receives OB115 and Solid G4 Mini Inverter data on RS485 Modbus Input
//It drives a 2004 I2C LCD display
//Data values are re-transmitted in Nextion Display format to either a remote Nextion display and or a remote Smart power socket

//NOTE MODBUS INPUT USES THE ARDUINO'S PRIMARY HARDWARE SERIAL PORT - So Dont use with Serial Monitor !!!!!!!!
//....SERIAL PRINT COMMANDS OR OTHER SERIAL COMMS CAN SCREW the MODBUS OPERATION!!!!!!!!!


#include <SPI.h>           // Required for I2C LCD
#include <Wire.h>          //  ""  ""
#include <LiquidCrystal_I2C.h>
#include <ModbusRtu.h>    // https://github.com/smarmengol/Modbus-Master-Slave-for-Arduino
#include <SoftwareSerial.h>
//#include <GyverDimmer.h>
//#include <GyverPID.h>


#define TXEN  3           //Pin connected to MAX 485 TXen/RXEn

// For Triac Driver 
//const byte ZeroCrossPin = 2;   // This must be an interrupt input pin, On Nano/Uno Atmel 328P devices D2==INT0, D3==INT1
//#define D_PIN 9

//DimmerBres<D_PIN> dim;   // specify dimmer pin


// For SoftwareSerial
SoftwareSerial mySerial(15, 17); // Rx=D15=A1, Tx=D17=A3 creating a 2nd Software based serial port - used for Remote Nextion Display output OR Wireless received data input


// For Rolling Ave
const int numReadings  = 20;  // Moving average window size - how many readings are averaged?
int AvgArray [numReadings];   // Setting up the array 
int readIndex  = 0;           // Where in the array the last values is saved
long total  = 0;              // the Sum of all values in the array
byte cvalues = 0;             // Count of values read into theRolling Ave function (<= nvalues), used to avoid readings rising at start up as array is empty
int TXmitCycle =0;


// For Modbus
uint16_t OB115data[14];   // data arrays for modbus network sharing
uint16_t SolisG4data[20];

uint8_t u8state; //!< machine state
uint8_t u8query; //!< pointer to message query


// Variables For Solis G4
long InverterPower = 0;
float PV_Volts = 0;
float PV_Amps =0;

// Variables for OB115
float ACVolts = 0;
float Frequency =0;
float ACCurrent = 0;
float ActPower = 0;
float ApparPower = 0;
float ReacPower = 0;
float PowerFactor = 0;

//General Program status Flags & Variables 
boolean LCDUpdated =false;
boolean NewData = false;
unsigned long msec_start = 0; 
unsigned long msec_in = 0;
unsigned long msec_out = 0;
unsigned long LastTXmit =0;
int TxInterval=0;
int TxCount =0; 

const byte ndt[3] = {255,255,255};  // Nextion commands are terminated by sending (0xff) 3 times

// For PID Controller in Integer 
int PowerNow = 0;   //used to hol int of ActPower
int PowerAve =0;


//For Software Serial Input reading

uint8_t rxByte;
uint8_t rxPtr = 0;
uint8_t rxBuf[40] = {0};
int32_t rxVal;
char* rxTxt;

//For Serial Data Parsing funct..
 char * strtokIndx; // this is used by strtok() as an index
 int RxdInt1 =0;


byte solar[8] =                //solar panel icon #1
{
  0b11111,
  0b10101,
  0b11111,
  0b10101,
  0b11111,
  0b10101,
  0b11111,
  0b00000,
};

byte energy[8] =               // Power flash #3
{
  0b00010,
  0b00100,
  0b01000,
  0b11111,
  0b00010,
  0b00100,
  0b01000,
  0b00000,
};

byte powerplug[8] =         // Power Plug Icon  #6
{
  0b01010,
  0b11111,
  0b10001,
  0b10001,
  0b10001,
  0b01110,
  0b00100,
  0b00100,
};


byte DegC[8] =         //Deg C in one char  #7
{
  0b01000,
  0b00000,
  0b00011,
  0b00100,
  0b00100,
  0b00100,
  0b00011,
  0b00000,
};


static char LCDstr[16];    //String array declaration, used to format Float values for display on LCD

LiquidCrystal_I2C lcd (0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);
Modbus master(0, 0, TXEN);
modbus_t telegram[2];

unsigned long u32wait;


//For PID
//GyverPID regulator(0.20, 0.09, 0.01, 200);  // coeff. P, cal. And, cal. D, sampling period dt (ms)
// or like this:
// GyverPID regulator(0.1, 0.05, 0.01);  // you can P, I, D, without dt, dt will be by default. 100 ms



//-------------------------------------------------------------------------
void setup() {

  lcd.begin(20, 4);
  lcd.setCursor(1, 0);
  lcd.print("OB115 Energy Meter");
  lcd.setCursor(3, 1);
  lcd.print("Solis G4 Mini");
  lcd.setCursor(0, 2);
  lcd.print("Modbus RTU Interface");
  lcd.setCursor(0, 3);
  lcd.print("--------V2---------");

  delay(2000);
  
  lcd.createChar(1, solar);
  lcd.createChar(2, energy);
  lcd.createChar(3, powerplug);
  lcd.createChar(4, DegC);
   
  lcd.clear();

  

  telegram[0].u8id = 2;                       // slave address
  telegram[0].u8fct = 4;                      // function code
  telegram[0].u16RegAdd = 2;                  // start address in slave
  telegram[0].u16CoilsNo = 14;                // number of elements (coils or registers) to read
  telegram[0].au16reg = OB115data;             // pointer to a memory array in the Arduino

  telegram[1].u8id = 1;                      // slave address
  telegram[1].u8fct = 4;                     // function code
  telegram[1].u16RegAdd = 3005;              // start address in slave
  telegram[1].u16CoilsNo = 18;                // number of elements (coils or registers) to read
  telegram[1].au16reg = SolisG4data;           // pointer to a memory array in the Arduino


  mySerial.begin(38400);  // Start the SoftwareSerial Port
  Serial.begin(9600);    //  Start Hardware Serial port, 
  
  //Initialising ModBusRTU
  master.start();
  master.setTimeOut(90 ); // if there is no answer in n ms, roll over
 
  u32wait = millis() + 150;
  u8state = u8query = 0; 


// create an interrupt from the zero crossing detector
//attachInterrupt(0, isr, RISING);  // D2 == 0
//attachInterrupt (digitalPinToInterrupt (ZeroCrossPin),isr, RISING);  // attach interrupt handler 


//For PID
//  regulator.setDirection(NORMAL); // direction of regulation (NORMAL / REVERSE). DEFAULT COST NORMAL
//  regulator.setLimits(0, 255);    // limits (set for 8 bit PWM). DEFAULT COST 0 AND 255
//  regulator.setpoint = -10;        // tell the regulator the temperature that it should maintain

//  in the process of work, you can change the coefficients
//  regulator.Kp = .02;
//  regulator.Ki += .02;
//  regulator.Kd = 0;

  LCDLayout();  //This function just lays out text on LCD - loop only updates values for speed improvement

   msec_start=millis();
}

//_______________________________________________________________________________________________________________

void loop() {
 

 // Manage Modbus Activity:-
 
  ModbusInput ();    //Recieves serial data and parses into variables 
 // SerialInput();
  
  if (NewData){
    // PIDController();    //Call PID update function
      
      PowerNow = (int) ActPower;       // Casting the float ActPower value to an Int
      PowerAve=RollingAve(PowerNow);   // Power now goes into a function which returns the computed Rolling Average power for display pursposes

      TXmitData();         //Function to Send data values over Software Serial Port  NOT USED IF SERIAL INPUT MODE IS USED
  
      TxRate();            //function determining average transmission interval in mS based on 10 transmits
         
      u32wait = millis() + 2; //This will start a new Modbus cycle
     
      NewData=false;
  }
   
}



//***************************************************************************************
void ModbusInput (){

switch( u8state ) {
  case 0: 
     if (millis() > u32wait){      // Detect the Modbus Scan interval time out, start a Modbus Query & Data collect sequence
     u8state++;
     
    }
    break;
  
  case 1: 
    
    
    master.query( telegram[0] ); // Send Modbus Query (only once)
    u8state++;
   break;
     
  case 2:
    master.poll(); // Now check for incoming data messages from slave
   
    if (master.getState() == COM_IDLE) {    //Umm..not sure if there should be some delay between poling data then reading registers??
      
    ACVolts = Modbus_U16Pair_to_Float(OB115data[0], OB115data[1] );    // Call function to reconstruct the float from 2 unsigned integers 
    Frequency = Modbus_U16Pair_to_Float(OB115data[2], OB115data[3] );    // Call function to reconstruct the float from 2 unsigned integers
    ACCurrent = Modbus_U16Pair_to_Float(OB115data[4], OB115data[5] );    // Call function to reconstruct the float from 2 unsigned integers 
    ActPower = Modbus_U16Pair_to_Float(OB115data[6], OB115data[7] );    // Call function to reconstruct the float from 2 unsigned integers 
    ApparPower = Modbus_U16Pair_to_Float(OB115data[8], OB115data[9] );    // Call function to reconstruct the float from 2 unsigned integers 
    ReacPower = Modbus_U16Pair_to_Float(OB115data[10], OB115data[11] );    // Call function to reconstruct the float from 2 unsigned integers 
    PowerFactor = Modbus_U16Pair_to_Float(OB115data[12], OB115data[13] );    // Call function to reconstruct the float from 2 unsigned integers
  
 //from query slave 0 to here =51ms
  
   u8state++;
   }
   else if(!LCDUpdated){
   UpdateLCDData();     //Updating data numbers only =41mS
   LCDUpdated=true;
   }
   break;

  case 3: 
    LCDUpdated=false;
    master.query( telegram[1] ); // Send Modbus Query (only once)
    u8state++;
   break;
    
  case 4:
    master.poll(); // check incoming Modbus data messages from the next slave
   
    if (master.getState() == COM_IDLE) {
     //Unpack registers for Solis data
     InverterPower = ModbusU16PairtoU32(SolisG4data[0],SolisG4data[1]); // Function to reconstruct 16bit register pair @address 3005-6 to Unsigned 32) 
     PV_Volts = SolisG4data[16]/10.0;    //PVVolts Reg address 3021
     PV_Amps = SolisG4data[17]/10.0;    //PVAmps Reg address
   
   
    u8state++;
    }
    break;

    
    case 5:           //Modbus reading cycle is now completed
    
    NewData=true;
    u8state = 0;    
   // u32wait = millis() + 2; 
     
    break; 
   }            // End of switch statement

}
 
 // Serial.print("looptime= ");
 // Serial.println(millis()-msec_in);    //For timing testing

//*************************************************************************************


// **************************************************************************************************
void SerialInput (){
  
  if (mySerial.available()) {
    rxByte = mySerial.read();
    rxBuf[rxPtr++] = rxByte; // add the byte to the buffer only if there is enough space left
  
    if ((*(uint32_t*)(rxBuf + rxPtr - 4) & 0xFFFFFF00) == 0xFFFFFF00) { // check for 3x0xFF....we have found the nextions end of message ....
      
        rxBuf[rxPtr - 3] = '\0'; //change the Nextion terminator into a c_string terminator
        rxTxt = (char*)rxBuf ; //take the buffer from the 2nd byte to eliminate the 0x70
       // Serial.println(rxTxt);
        ParseData();
        NewData=true;
        rxPtr = 0;
        //Serial.println("***");
    }
  }
}

//***************************************************************************************************
 void ParseData () {
    
     char* token;           //array of Chars
     char* VariableID ;     //

 
    strtokIndx = strtok(rxTxt, "=");  // get the first part of string before the first space Char
    VariableID = strtokIndx;
   // Serial.println(strtokIndx);

        
    strtokIndx = strtok(NULL, " ");  // get the remaining part of string after the "=" ....which is the variables int number 
    RxdInt1 = atoi(strtokIndx);
       
    //Serial.println(RxdInt1);

    if(strcmp(VariableID, "ActPow.val") == 0) {
    // Serial.println("Found variable ActPower.val");
     ActPower = RxdInt1;
    }
    else if (strcmp(VariableID, "VAC.val") == 0) {
    // Serial.println("Found variable VAC.val");
     ACVolts = RxdInt1;
     }
    else if (strcmp(VariableID, "InvPow.val") == 0) {
    // Serial.println("Found variable InvPow.val");
     InverterPower = RxdInt1;
    }
    else if (strcmp(VariableID, "PVVolt.val") == 0) {
    // Serial.println("Found variable InvPow.val");
     PV_Volts = RxdInt1;
    }
    else if (strcmp(VariableID, "PVAmp.val") == 0) {
    // Serial.println("Found variable InvPow.val");
     PV_Amps = RxdInt1;
    }
 }


//***************************************************************************************
void isr() {        //Interrupt calls for update on Triac driving routine
//  dim.tick(); // trigger a tick in the zero detector interrupt
}

//****************************************************************************************
void  PIDController(void){       //Pass vlues into a PID control algorithm and use the output value to set Triac Dimmer trigger level

PowerNow = (int) ActPower;       // Casting the float ActPower value to an Int

PowerAve=RollingAve(PowerNow);   // Power now goes into a function which returns the computed Rolling Average power for display pursposes

//regulator.input = PowerNow;      // informing the PID regulator of the current Power 

//int dimmerVal = regulator.getResultTimer();   //Gets the next output drive value from the PID algorithm

//dim.write(dimmerVal);  // Dimmer setting range is 0-255

//Serial.print("dimVal=");
//Serial.println(dimmerVal);
 
// delay(100);

}



 // *************************************************************
long RollingAve(int value) {      
  
  total+= value;          

  if (cvalues == numReadings)    // If the window is full, adjust the sum by deleting the oldest value
    total -= AvgArray[readIndex];

  AvgArray[readIndex] = value;          // Replace the oldest with the latest

  if (++readIndex >= numReadings)

    readIndex = 0;

  if (cvalues < numReadings)
    cvalues += 1;               //equivalent to the expression x = x + y  !, Once cvalues reaches numReadings it no longer increments.
                                // Average thus starts as total/1 after first values is received, then Total/2 on second....later it becomes Total/ numReadings

  return total/cvalues;
}






//*********************

void LCDLayout(void)
{

  
  lcd.setCursor(0,0);
  lcd.write(1);     //PV symbol
  lcd.print("Array:");
  lcd.setCursor(13,0);
  lcd.print("V/");
  
  
  lcd.setCursor(19,0);
  lcd.print("A");


  lcd.setCursor(0, 1);
  lcd.write(3);
  lcd.print("Inverter:");
  lcd.setCursor(19, 1);
  lcd.print("W");
 
  lcd.setCursor(0, 2);
  lcd.write(2);
  lcd.print ("Grid ");
  
  lcd.setCursor(19, 2);
  lcd.print("W");
    
  lcd.setCursor(0, 3);
  //lcd.write(2);
  lcd.print("VAC:");
    
  lcd.setCursor(9, 3);
  lcd.print("V ");
  
   
  lcd.setCursor(19, 3);
  lcd.print("W");

}
//*************************************************************************
void TxRate (void)
{
TxCount++;
  if (TxCount >=10){
    TxInterval = ((millis()-msec_start)/10);
   
    mySerial.print(F("TxIntv.val="));
    mySerial.print((int)TxInterval);      // Tranmit ActPower Variable (from OB115 Meter) to Nextion ,+ve values are import, -Ve for export
    mySerial.write(ndt,3);              // command terminated by sending (0xff) 3 times
  
   msec_start=millis();
   TxCount =0;
  }
}
//*************************************************************************
void UpdateLCDData(void){
  
  dtostrf(PV_Volts, 5, 1, LCDstr);
  lcd.setCursor(8,0);
  lcd.print(LCDstr);
  
  
  dtostrf(PV_Amps, 5, 1, LCDstr);
  lcd.setCursor(14,0);
  lcd.print(LCDstr);

  

  dtostrf(InverterPower, 5,0, LCDstr);
  lcd.setCursor(11, 1);
  lcd.print(LCDstr);
  
  dtostrf(ActPower, 5, 0, LCDstr);
  lcd.setCursor(6, 2);
      if(ActPower>0.1){
        lcd.print ("Import");
      }
      else if(ActPower < -0.1){
        lcd.print ("Export");
      }
   //   else{
  //      lcd.print ("      ");
   //   }
  lcd.setCursor(14, 2);
  lcd.print(LCDstr);
  
  dtostrf(ACVolts, 5,1, LCDstr);   
  lcd.setCursor(4, 3);
  lcd.print(LCDstr);
  
  dtostrf(PowerAve, 5,0, LCDstr);
  lcd.setCursor(13, 3);
  lcd.print(LCDstr);  //Temporarily this is the rolling ave of power export / i,[port
 

}



//**********************************************************************
float Modbus_U16Pair_to_Float(uint16_t a, uint16_t b) {
    uint32_t combined = ((uint32_t)a << 16) | b;
    float f;
    memcpy(&f, &combined, sizeof f);
    return f;
}

//**********************************************************************

long ModbusU16PairtoU32(uint16_t a, uint16_t b) {
   long f = a + (b * 65536) ;
   return f;
 }

 //*********************************************************************
